import os
import torch
from typing import Dict, Any, Optional, Self
from confuse import Configuration
import numpy as np
from scipy.fftpack import idct

ORTHOGONAL_PROJECTION_NAME = 'orthogonal_projection'
ZERO_MEAN_PROJECTION_NAME = 'zero_mean_projection'
UNIT_BALL_PROJECTION = 'unit_ball_projection'

def unit_ball_projection(x: torch.Tensor) -> torch.Tensor:
    return x / torch.sum(torch.abs(x), dim=(-2, -1), keepdim=True)

def zero_mean_projection(x: torch.Tensor) -> torch.Tensor:
    return x - torch.mean(x, dim=(-2, -1), keepdim=True)

def orthogonal_projection_procrustes(x: torch.Tensor, eps: float=1e-7, max_num_iterations: int=5,
                                     rel_tol: float=1e-5) -> torch.Tensor:
    """
    Function which applies the orthogonal procrustes scheme to orthogonalise a set of filters

    :param x: Filter tensor of shape [batch_size, 1, filter_dim, filter_dim]
    :param eps: Clipping parameter to ensure positivity of diagonal elements
    :param max_num_iterations: Maximal number of iterations
    :param rel_tol: Tolerance used to stop iteration as soon the norm of subsequent iterates is less than rel_tol.
    :return: Orthogonalised set of filters in terms of a tensor of the same shape as the input tensor.
    """
    x_flattened = [x[i, 0, :, :].flatten() for i in range(0, x.shape[0])]
    x_stacked = torch.stack(x_flattened, dim=1)
    m, n = x_stacked.shape
    diag = torch.diag(torch.ones(n, dtype=x.dtype, device=x.device))

    v_old = torch.zeros_like(x_stacked)
    for k in range(0, max_num_iterations):
        z = torch.matmul(x_stacked, diag)
        q, s, r_h = torch.linalg.svd(z, full_matrices=False)
        v = torch.matmul(q, r_h)

        tmp = torch.matmul(v.transpose(dim0=0, dim1=1), x_stacked)
        diag_elements = torch.diag(tmp)
        diag_elements = torch.clamp(diag_elements, min=eps)
        diag = torch.diag(diag_elements)
        if torch.linalg.norm(v - v_old) < rel_tol:
            break
        v_old = v.clone()

    x_orthogonal = [torch.unflatten(torch.matmul(v, diag)[:, j], dim=0,
                                    sizes=x.shape[-2:]) for j in range(0, v.shape[1])]
    return torch.stack(x_orthogonal, dim=0).unsqueeze(dim=1)

class ImageFilter(torch.nn.Module):
    """
    Class modelling quadratic image filters by means of a PyTorch module.

    NOTE
    ----
        > Instance of this class can be created by means of constructor, or
            by means of the class methods
                * from_file()
                * from_config()
            The latter requires as parameter a configuration object from the
            Python package confuse.
    """
    def __init__(
            self, 
            filter_dim: int=7, 
            padding: int=3, 
            padding_mode: str='reflect', 
            init_options: Optional[Dict[str, int]]=None,
            orthogonality_options: Optional[Dict[str, int]]=None,
            apply_unit_ball_projection: bool=False,
            trainable: bool=True, 
    ) -> None:

        """
        Initialisation of class ImageFilter.

        TODO
        ----
            > Introduce backward-hooks for proper handling of projections. Currently these
                are applied 'manually' in the optimisation routines.

        :param filter_dim: Filter dimension - 7 by default.
        :param padding: Integer giving the padding which is used when applying filters; 3 by default.
        :param padding_mode: String indicating the padding mode; 'reflect' by default.
        :param init_options: Dictionary collecting options for the initialisation of image filters. Keys:
            - 'mode': String indicating how filters are initialised: 'rand', 'randn', 'dct'. The latter
                corresponds to the initialisation of the filters by means of the discrete cosine transform. By default
                'rand' is used.
            - 'multiplier': Float used as scaling factor in the initialisation process; default value equals 1
            - 'normalise': Flag indicating if filters are normalised w.r.t. l2 norm within the initialisation procedure. 
                By default normalisation is not applied. 
        :param orthogonality_options: Dictionary with keys 'enable', 'max_num_iterations'. If orthonalisation
            is enabled, the procrustes algorithm is used to orthogonalise filters at initialisation stage and
            after each update step. For this purpose a callable attribute titled ORTHOGONAL_PROJECTION_NAME
            is created.
            By default orthogonalisation is not applied. Default number of procrustes iterations equals to 5.
        :param apply_unit_ball_projection: Flag indicating if filters are projected onto unit ball w.r.t.
            1-norm.
        :param trainable: Flag indicating if filter is trainable or not. True by default.
        """
        super().__init__()

        default_init = {
            'mode': 'rand', 
            'multiplier': 1.0, 
            'normalise': False
        }
        self.init_options = {**default_init, **(init_options or {})}

        self.filter_dim = filter_dim
        self.padding = padding
        self.padding_mode = padding_mode
        self.orthogonality_options = orthogonality_options if orthogonality_options is not None else {}
        self.apply_unit_ball_projection = apply_unit_ball_projection

        # initialisation with constructor parameters
        filter_data = self._init_filter_tensor(filter_dim, self.init_options,
                                               self.orthogonality_options)
        self.filter_tensor = torch.nn.Parameter(data=filter_data, requires_grad=trainable)
        self._bind_projections()

    def _bind_projections(self) -> None:
        """
        Binding of callable attributes to self.filter_tensor. Orthogonal procrustes projection is
        not bound by default.
        """
        if not hasattr(self.filter_tensor, ZERO_MEAN_PROJECTION_NAME):
            setattr(self.filter_tensor, ZERO_MEAN_PROJECTION_NAME, zero_mean_projection)

        orthogonalise = self.orthogonality_options.get('enable', False)
        if orthogonalise:
            max_num_procrustes_iteations = self.orthogonality_options.get('max_num_iterations', 5)
            def orthogonal_projection(x: torch.Tensor):
                return orthogonal_projection_procrustes(x, max_num_iterations=max_num_procrustes_iteations)

            setattr(self.filter_tensor, ORTHOGONAL_PROJECTION_NAME, orthogonal_projection)

        if self.apply_unit_ball_projection and not hasattr(self.filter_tensor, UNIT_BALL_PROJECTION):
            setattr(self.filter_tensor, UNIT_BALL_PROJECTION, unit_ball_projection)

    @staticmethod
    def _init_filter_tensor(
            filter_dim: int, 
            init_options: Dict[str, Any],
            orthogonality_options: Dict[str, Any],
    ) -> torch.Tensor:
        """
        Utiliy method for the initialisation of the parameter filter_tensor.

        :param filter_dim:
        :param initialisation_mode:
        :param normalise:
        :param orthogonality_options:
        :param multiplier:
        :return: PyTorch tensor of shape [num_filters, 1, filter_dim, filter_dim]
        """
        if init_options['mode'] == 'dct':
            can_basis = np.reshape(np.eye(filter_dim ** 2, dtype=np.float64),
                                    (filter_dim ** 2, filter_dim, filter_dim))
            dct_basis = idct(idct(can_basis, axis=1, norm='ortho'), axis=2, norm='ortho')
            dct_basis = dct_basis[1:].reshape(-1, 1, filter_dim, filter_dim)
            filter_data = torch.tensor(dct_basis)
        elif init_options['mode'] == 'randn':
            filter_data = torch.randn(filter_dim ** 2 - 1, 1, filter_dim, filter_dim)
        elif init_options['mode'] == 'rand':
            filter_data = 2 * torch.rand(filter_dim ** 2 - 1, 1, filter_dim, filter_dim) - 1
        else:
            raise ValueError('Unknown initialisation method.')

        orthogonalise = orthogonality_options.get('enable', False)
        if orthogonalise:
            max_num_procrustes_iteations = orthogonality_options.get('max_num_iterations', 5)
            orthogonal_projection_procrustes(filter_data, max_num_iterations=max_num_procrustes_iteations)
        if init_options['normalise']:
            filter_data.divide_(torch.linalg.norm(filter_data, dim=(-2, -1)).reshape(-1, 1, 1, 1))
        filter_data = zero_mean_projection(filter_data)
        filter_data.mul_(init_options['multiplier'])

        return filter_data

    @classmethod
    def from_config(cls, config: Configuration) -> Self:
        """
        Class method for initialisation from config.

        :param config: Configuration object from Python package confuse.
        :return: Instance of class ImageFilter
        """
        filter_dim = config['image_filter']['filter_dim'].get()
        padding = config['image_filter']['padding'].get()
        padding_mode = config['image_filter']['padding_mode'].get()

        initialisation_mode = config['image_filter']['initialisation']['mode'].get()
        multiplier = config['image_filter']['initialisation']['multiplier'].get()
        normalise = config['image_filter']['initialisation']['normalise'].get()
        trainable = config['image_filter']['trainable'].get()

        orthogonality_options = {
            'enable': config['image_filter']['orthogonality']['enable'].get(),
            'max_num_iterations': config['image_filter']['orthogonality']['max_num_iterations'].get()
        }
        init_options = {'mode': initialisation_mode, 'multiplier': multiplier, 'normalise': normalise}
        apply_unit_ball_projection = config['image_filter']['apply_unit_ball_projection'].get()
        
        return cls(filter_dim, padding, padding_mode, init_options, 
                   orthogonality_options, apply_unit_ball_projection, trainable)

    @classmethod
    def from_file(cls, path_to_model: str, device: torch.device=torch.device('cpu')) -> Self:
        """
        Class method for initialisation from file.

        :param path_to_model: String representing the path to the model file.
        :param device: Location where loaded model should be placed; by default model is placed on cpu.
        :return: Instance of class ImageFilter
        """
        filter_data = torch.load(path_to_model, map_location=device)

        initialisation_dict = filter_data['initialisation_dict']
        state_dict = filter_data['state_dict']

        filter_dim = initialisation_dict.get('filter_dim', 7)
        padding = initialisation_dict.get('padding', 3)
        padding_mode = initialisation_dict.get('padding_mode', 'reflect')
        orthogonality_options = initialisation_dict.get('orthogonality_options', {})
        apply_unit_ball_projection = initialisation_dict.get('apply_unit_ball_projection', False)

        image_filter = cls(filter_dim=filter_dim,
                           padding=padding,
                           padding_mode=padding_mode,
                           orthogonality_options=orthogonality_options, 
                           apply_unit_ball_projection=apply_unit_ball_projection)
        image_filter.load_state_dict(state_dict, strict=True)
        return image_filter

    def get_filter_tensor(self) -> torch.Tensor:
        return self.filter_tensor.data

    def get_num_filters(self) -> int:
        return self.filter_tensor.shape[0]

    def freeze(self) -> None:
        self.filter_tensor.requires_grad = False

    def unfreeze(self) -> None:
        self.filter_tensor.requires_grad = True

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_padded = torch.nn.functional.pad(x, (self.padding, self.padding,
                                               self.padding, self.padding), self.padding_mode)
        return torch.nn.functional.conv2d(x_padded, self.filter_tensor)

    def state_dict(self, *args, **kwargs) -> Dict[str, Any]:
        state = super().state_dict(*args, **kwargs)
        return state

    def initialisation_dict(self)  -> Dict[str, Any]:
        return {'filter_dim': self.filter_dim,
                'padding': self.padding,
                'padding_mode': self.padding_mode,
                'apply_unit_ball_projection': self.apply_unit_ball_projection,
                'orthogonality_options': self.orthogonality_options}

    def save(self, path_to_model_dir: str, model_name: str='filters') -> str:
        path_to_model = os.path.join(path_to_model_dir, '{:s}.pt'.format(os.path.splitext(model_name)[0]))
        filter_data_dict = {'initialisation_dict': self.initialisation_dict(),
                            'state_dict': self.state_dict()}

        torch.save(filter_data_dict, path_to_model)
        return path_to_model



import torch

def poly_func(x):

    a = 3
    b = 2
    c = 1
    d = 4
    return a + x * (b + x * (c + x * d))

def main():

    x = torch.rand(1)
    y = poly_func(x)


    xx = torch.ones(2).requires_grad_(True)
    vv = 2 * torch.ones_like(xx)
    with torch.enable_grad():
        yy = poly_func(xx)

    dy_dx = torch.autograd.grad(inputs=xx, outputs=yy, grad_outputs=vv)

    print(yy)
    print(dy_dx)

if __name__ == '__main__':
    main()



from typing import Any, Tuple

import torch

def cube_forward(x):
    return x**3

def cube_backward(grad_out, x):
    return grad_out * 3 * x**2

def cube_backward_backward(grad_out, sav_grad_out, x):
    return grad_out * sav_grad_out * 6 * x

def cube_backward_backward_grad_out(grad_out, x):
    return grad_out * 3 * x**2

class Cube(torch.autograd.Function):
    @staticmethod
    def forward(x, s='grrrrrr'):
        t = 'asd'
        return cube_forward(x), t

    @staticmethod
    def setup_context(ctx: Any, inputs: Tuple[Any, ...], output: Any) -> Any:
        x, s = inputs
        _, t = output
        ctx.save_for_backward(x)
        ctx.s = s
        ctx.t = t

    @staticmethod
    def backward(ctx, *grad_out):
        x, = ctx.saved_tensors
        s = ctx.s
        return CubeBackward.apply(grad_out[0], x), None

class CubeBackward(torch.autograd.Function):
    @staticmethod
    def forward(grad_out, x):
        return cube_backward(grad_out, x)

    @staticmethod
    def setup_context(ctx: Any, inputs: Tuple[Any, ...], output: Any) -> Any:
        grad_out, x = inputs
        ctx.save_for_backward(x)
        ctx.grad_out = grad_out

    @staticmethod
    def backward(ctx, *grad_out):
        x, = ctx.saved_tensors
        sav_grad_out = ctx.grad_out
        dx = cube_backward_backward(grad_out[0], sav_grad_out, x)
        dgrad_out = cube_backward_backward_grad_out(grad_out[0], x)
        return dgrad_out, dx


x = torch.tensor(3., requires_grad=True, dtype=torch.double)

with torch.enable_grad():
    y = Cube.apply(x, 'bla')[0]
    dy_dx = torch.autograd.grad(inputs=x, outputs=y, create_graph=True)[0].requires_grad_(True)
print(dy_dx)
print(dy_dx.requires_grad)

v = torch.tensor(-6.66)
d2y_dx2 = torch.autograd.grad(inputs=x, outputs=dy_dx, grad_outputs=v)

print('asd')

def func(z):
    return Cube.apply(z, 'yippie')[0]

torch.autograd.gradcheck(func, (x,))
torch.autograd.gradgradcheck(func, (x, ))


import torch

from bilevel_optimisation.filters import ImageFilter
from bilevel_optimisation.potential import StudentT, Spline, Potential
from bilevel_optimisation.fields_of_experts import FieldsOfExperts
from bilevel_optimisation.utils.config_utils import load_app_config

def second_derivative_filter(config):
    image_filter = ImageFilter(config)

    x = torch.rand(16, 1, 64, 64).requires_grad_(True)
    y = torch.sum(image_filter(x))

    dy_dx = torch.autograd.grad(inputs=x, outputs=y, create_graph=True)
    d2y_d2x = torch.autograd.grad(inputs=x, outputs=dy_dx, grad_outputs=torch.ones_like(x))[0]

    torch.allclose(d2y_d2x, torch.zeros_like(x))

def second_derivative_regulariser(config):
    x = torch.rand(4, 1, 16, 16).requires_grad_(True)

    image_filter = ImageFilter(config)
    potential = Potential.from_config(image_filter.get_num_filters(), config)
    regulariser = FieldsOfExperts(potential, image_filter)

    v = 7 * torch.ones_like(x)


    # if potential.__class__.__name__ == 'StudentT':
    # y = regulariser(x)
    #     dy_dx = torch.autograd.grad(inputs=x, outputs=y, create_graph=True)[0]
    #     d2y_dx2 = torch.autograd.grad(inputs=x, outputs=dy_dx, grad_outputs=v)[0]
    #     u = regulariser.image_filter(x).detach().requires_grad_(True)
    #     p = regulariser.potential.forward_negative_log(u)
    #     dp_du = torch.autograd.grad(inputs=u, outputs=p, create_graph=True)
    #     d2p_du2 = torch.autograd.grad(inputs=u, outputs=dp_du, grad_outputs=image_filter(v))[0]
    #
    #     print(torch.linalg.norm(d2y_dx2 - d2p_du2))

    if potential.__class__.__name__ == 'NaturalCubicSpline':

        # # first potential derivatives
        # y = potential.forward_negative_log(x)
        # dy_dx = torch.autograd.grad(inputs=x, outputs=y)[0]
        # dy_dx_ = potential.first_derivative(x)
        # print(torch.linalg.norm(dy_dx_ - dy_dx))
        #
        # # first regulariser derivatives
        # z_inter = image_filter(x).detach().requires_grad_(True)
        # dz_dx = torch.autograd.grad(inputs=z_inter, outputs=potential.forward_negative_log(z_inter))[0]
        # dz_dx_ = potential.first_derivative(image_filter(x))
        # print(torch.linalg.norm(dz_dx_ - dz_dx))
        #
        # dr_dx = torch.autograd.grad(inputs=x, outputs=regulariser(x), grad_outputs=)[0]
        # dr_dx_ = potential.first_derivative(image_filter(x)) * torch.autograd.grad(inputs=x, outputs=torch.sum(image_filter(x)))[0]

        # grad_outs = torch.autograd.grad(inputs=x, outputs=image_filter(x), grad_outputs=v)[0]
        # hvp = potential.second_derivative(x, grad_outputs=grad_outs, mixed=False)[0]
        # hvp_ = torch.autograd.grad(inputs=x, outputs=image_filter(x), grad_outputs=hvp)

        y = potential.forward_negative_log(x)
        y1 = potential.first_derivative(x)

        y2 = potential.second_derivative(x, grad_outputs=v)

        yy = torch.autograd.grad(inputs=x, outputs=y, create_graph=True)[0]

        yyy = torch.autograd.grad(inputs=x, outputs=yy, grad_outputs=v)

        from torch.autograd import gradcheck, gradgradcheck

        def func(xx):
            return potential.forward_negative_log(xx)

        good_1st = torch.autograd.gradcheck(func, (x.to(torch.float64),))

        good_2nd = torch.autograd.gradgradcheck(func, (x.to(torch.float64),))




        print('ahoi')

        # print(torch.linalg.norm(dr_dx_ - dr_dx))

        # dy_dx = potential.first_derivative(x)




        d2y_dx2 = torch.autograd.grad(inputs=x, outputs=dy_dx, grad_outputs=v)[0]

        u = regulariser.image_filter(x).detach().requires_grad_(True)

        d2p_du2 = potential.second_derivative(u, image_filter(v))
        print(torch.linalg.norm(d2y_dx2 - d2p_du2))




def main():
    path_to_config_dir = '/home/florianthaler/Documents/research/bilevel_optimisation/bilevel_optimisation/config_data/custom/example_training_III'
    config = load_app_config('test', path_to_config_dir, 'test_module')

    # second_derivative_filter(config)
    second_derivative_regulariser(config)

if __name__ == '__main__':
    main()





import torch

def cube_forward(x, a):
    return a * x**3

def cube_backward(grad_out, x, a):
    return grad_out * 3 * a * x**2

def cube_backward_backward(grad_out, sav_grad_out, x, a):
    return grad_out * sav_grad_out * 6 * a * x

def cube_backward_backward_grad_out(grad_out, x, a):
    return grad_out * 3 * a * x**2

class Cube(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x):
        ctx.save_for_backward(x)
        return cube_forward(x, torch.ones(1))

    @staticmethod
    def backward(ctx, grad_out):
        x, = ctx.saved_tensors
        return CubeBackward.apply(grad_out, x)

class CubeBackward(torch.autograd.Function):
    @staticmethod
    def forward(ctx, grad_out, x):
        ctx.save_for_backward(x, grad_out)
        return cube_backward(grad_out, x, torch.ones(1))

    @staticmethod
    def backward(ctx, grad_out):
        x, sav_grad_out = ctx.saved_tensors
        dx = cube_backward_backward(grad_out, sav_grad_out, x, torch.ones(1))
        dgrad_out = cube_backward_backward_grad_out(grad_out, x, torch.ones(1))
        return dgrad_out, dx

class ParamCube(torch.autograd.Function):
    @staticmethod
    def forward(ctx, a, x):
        ctx.save_for_backward(x, a)
        return cube_forward(x, a)

    @staticmethod
    def backward(ctx, grad_out):
        x, a = ctx.saved_tensors
        return None, ParamCubeBackward.apply(grad_out, a, x)

class ParamCubeBackward(torch.autograd.Function):
    @staticmethod
    def forward(ctx, grad_out, a, x):
        ctx.save_for_backward(x, a, grad_out)
        return cube_backward(grad_out, x, a)

    @staticmethod
    def backward(ctx, grad_out):
        x, a, sav_grad_out = ctx.saved_tensors
        dx = cube_backward_backward(grad_out, sav_grad_out, x, a)
        dgrad_out = cube_backward_backward_grad_out(grad_out, x, a)
        return None, dgrad_out, dx

def main():
    x = torch.tensor(1., requires_grad=True)
    a = -torch.ones(1)

    with torch.enable_grad():
        y = Cube.apply(x)
        dy_dx = torch.autograd.grad(inputs=x, outputs=y, create_graph=True)
    d2y_d2x =  torch.autograd.grad(inputs=x, outputs=dy_dx, create_graph=True)

    print('########################################################################')

    with torch.enable_grad():
        y = ParamCube.apply(x, a)
        dy_dx = torch.autograd.grad(inputs=x, outputs=y, create_graph=True)
    d2y_d2x =  torch.autograd.grad(inputs=x, outputs=dy_dx, create_graph=True)

    print(y)
    print(dy_dx)
    print(d2y_d2x)

if __name__ == '__main__':


    main()

# torch.autograd.gradcheck(Cube.apply, x)
# torch.autograd.gradgradcheck(Cube.apply, x)


import torch

torch.manual_seed(123)

# Dummy parameters: two independent 1x1xWxH tensors
w, h = 4, 4

x = torch.randn(1, 1, w, h)
y = torch.randn(1, 1, w, h)

p1_1 = torch.nn.Parameter(x.detach().clone())
p2_1 = torch.nn.Parameter(y.detach().clone())


param_groups_1 = [
    {'params': [p1_1], 'lr': 0.1},
    {'params': [p2_1], 'lr': 0.1},
]

for _ in range(0, 3):

    loss_1 = torch.sum(p1_1 ** 2 + p2_1 ** 2)
    loss_1.backward()  # gradients computed once per leaf

    # Vanilla SGD step, applied per group
    with torch.no_grad():
        for group in param_groups_1:
            lr = group['lr']
            for p in group['params']:
                if p.grad is not None:
                    p -= lr * p.grad

    # Zero gradients for next iteration
    for group in param_groups_1:
        for p in group['params']:
            if p.grad is not None:
                p.grad.zero_()

print("p1:", p1_1)
print("p2:", p2_1)

p1_2 = torch.nn.Parameter(x.detach().clone())
p2_2 = torch.nn.Parameter(y.detach().clone())

param_groups_2 = [
    {'params': [p1_2], 'lr': 0.1},
    {'params': [p2_2], 'lr': 0.1},
]

optim = torch.optim.SGD(param_groups_2, lr=0.1)

for _ in range(0, 3):

    loss_2 = torch.sum(p1_2 ** 2 + p2_2 ** 2)
    loss_2.backward() 
    optim.step()
    optim.zero_grad()

print("p1:", p1_2)
print("p2:", p2_2)

p = torch.cat([x, y], dim=0).requires_grad_(True)
param_groups_3 = [{'params': [p]}]

optim = torch.optim.SGD(param_groups_3, lr=0.1)

for _ in range(0, 3):

    loss_3 = torch.sum(p ** 2)
    loss_3.backward() 
    optim.step()
    optim.zero_grad()

print(p)

p0 = torch.cat([x, y], dim=0).requires_grad_(True)
p0_1 = p0[0:1, :, :, :].detach().clone().requires_grad_(True)
p0_2 = p0[1:2, :, :, :].detach().clone().requires_grad_(True)
param_groups_4 = [
    {'params': [p0_1], 'lr': 0.1},
    {'params': [p0_2], 'lr': 0.1},
]

optim = torch.optim.SGD(param_groups_4, lr=0.1)
for _ in range(0, 3):
    loss_4 = torch.sum(p0_1 ** 2 + p0_2 ** 2)
    loss_4.backward() 
    optim.step()
    optim.zero_grad()

print("p1:", p0_1)
print("p2:", p0_2)


from abc import ABC, abstractmethod
from typing import Callable
import torch

from pylopt.data import SolverResult

class IterativeLinearSystemSolver(ABC):

    def __init__(self) -> None:
        pass

    @abstractmethod
    def solve(self, linear_operator: Callable, b: torch.Tensor, x0: torch.Tensor = None) -> SolverResult:
        pass


import os
import torch
from torch.utils.data import IterableDataset, Dataset
from torchvision.transforms import v2
from PIL import Image
import random
from abc import ABC

class ImageDataset(ABC):
    """
    Class, to be used as base dataset class, which models a dataset of images, where images
    are loaded from file. All images are transformed to grayscale images and
    represented as torch tensors of the specified data type (by default float64)
    """
    def __init__(self, root_path: str, dtype: torch.dtype = torch.float64) -> None:
        """
        Initialisation of class ImageDataset

        :param root_path: Path to the directory containing the images
        :param dtype: Datatype which is used to represent images in terms of tensors.
        """
        self._root_path = root_path
        self._image_list = os.listdir(root_path)
        self._transform = v2.Compose([v2.ToImage(), v2.ToDtype(dtype, scale=True),
                                      v2.Grayscale(num_output_channels=1)])

    def _load_image(self, image_name: str) -> Image.Image:
        return Image.open(os.path.join(self._root_path, image_name))

class TestImageDataset(ImageDataset, Dataset):
    """
    Class, inheriting from ImageDataset and torch.utils.data.Dataset, which is used to model
    a dataset specifically for testing.
    """

    def __init__(self, root_path: str, dtype: torch.dtype = torch.float64):
        ImageDataset.__init__(self, root_path, dtype)
        Dataset.__init__(self)

    def __len__(self):
        return len(self._image_list)

    def __getitem__(self, idx: int):
        image_name = self._image_list[idx]
        image = self._load_image(image_name)
        return self._transform(image).unsqueeze(dim=0)

class TrainingImageDataset(ImageDataset, IterableDataset):
    """
    Class, inheriting from ImageDataset and torch.utils.data.IterableDataset, modelling an image dataset
    for the purpose of training: Images are loaded randomly from file - sampling with replacement. Note that
    crops will not be generated by methods of this class - use a data loader and a collate function instead.

    """
    def __init__(self, root_path: str, dtype: torch.dtype = torch.float64) -> None:
        """
        Initialisation of an object of class TrainingImageDataset

        :param root_path: Path to image directory
        :param dtype: torch datatype the images are represented with
        """
        ImageDataset.__init__(self, root_path, dtype)
        Dataset.__init__(self)

    def __iter__(self) -> torch.Tensor:
        while True:
            image_name = random.choice(self._image_list)
            image = self._load_image(image_name)
            yield self._transform(image).unsqueeze(dim=0)



import os
from confuse import Configuration
from typing import Optional
import torch
import logging
from importlib import resources

TYPE_DICT = {'float16': torch.float16, 'float32': torch.float32, 'float64': torch.float64}

def load_app_config(app_name: str, path_to_config_super_dir: str, custom_config_dir: str, configuring_module: str) -> Configuration:
    default_config_dir_path = os.path.join(path_to_config_super_dir, 'default')
    custom_config_dir_path = os.path.join(path_to_config_super_dir, custom_config_dir)
    config = Configuration(app_name)
    for dir_path in [default_config_dir_path, custom_config_dir_path]:
        for file in os.listdir(dir_path):
            logging.info('[{:s}] load configs from {:s} '
                         'in {:s}'.format(configuring_module.upper(), file, dir_path))
            config.set_file(os.path.join(dir_path, file))

    return config

def parse_datatype(config: Configuration) -> Optional[torch.dtype]:
    type_str = config['data']['type'].get()
    dtype = None
    if type_str in TYPE_DICT.keys():
        dtype = TYPE_DICT[type_str]
    return dtype



import torch
from typing import Optional
from confuse import Configuration

class MeasurementModel(torch.nn.Module):
    """
    A class, inherited from torch.nn.Module modeling the noisy measurement/observation of data.
    Observations are assumed to be a sum of a forward operator applied to the clean data and noise sampled from
    the multivariate normal distribution with zero mean and covariance matrix (noise_level ** 2) * I.
    """
    def __init__(self, 
                 u_clean: torch.Tensor, 
                 operator: Optional[torch.nn.Module]=None, 
                 noise_level: Optional[float]=None, 
                 config: Optional[Configuration]=None) -> None:
        """
        Initialisation of an object of class MeasurementModel. 

        NOTE
        ----
            > Either both, operator and noise_level, or config must be specified.

        :param u_clean: Tensor of shape [batch_size, channels, height, width]
        :param operator: PyTorch module representing the forward operator
        :param noise_level: Float in the interval (0, 1) representing the noise level
        :param config: Configuration object for the configuration of operator and noise level.
        """
        super().__init__()
        self.u_clean = torch.nn.Parameter(u_clean, requires_grad=False)

        if config is not None:
            # TODO:
            #   > extend search for operators to submodule (operators f.e.) such that custom forward operators can be used

            operator_name = config['measurement_model']['forward_operator'].get()
            self.operator = getattr(torch.nn, operator_name)()
            self.noise_level = config['measurement_model']['noise_level'].get()
        elif noise_level is not None and operator is not None:
            self.operator = operator
            self.noise_level = noise_level
        else:
            raise ValueError('Must provide config or both, operator and noise_level.')

        self.u_noisy = torch.nn.Parameter(self.make_noisy_observation(), requires_grad = False)

    def get_clean_data(self) -> torch.nn.Parameter:
        """
        Returns the clean data tensor to the caller

        :return: torch.Tensor
        """
        return self.u_clean

    def make_noisy_observation(self) -> torch.Tensor:
        obs_clean = self.operator(self.u_clean)
        return obs_clean + self.noise_level * torch.randn_like(obs_clean)

    def get_noisy_observation(self) -> torch.nn.Parameter:
        return self.u_noisy

    def forward(self, u: torch.Tensor,) -> torch.Tensor:
        """
        Computes the data fidelty of the input tensor u, i.e. the scaled 
        squared l2-norm of the difference between the input tensor 
        and a noisy observation. 

        :param u: Tensor of the same shape as clean or noisy data tensor.
        :return: Data fidelty of u
        """
        return 0.5 * torch.sum((u - self.u_noisy) ** 2) / self.noise_level ** 2





import torch
from typing import Any, Dict, Self, Optional
from confuse import Configuration

from . import Potential
from quartic_bspline_extension.functions import QuarticBSplineFunction

class QuarticBSpline(Potential):

    def __init__(self, 
                 num_marginals: int,
                 box_lower: float=-1.0,
                 box_upper: float=1.0,
                 num_centers: int=87,
                 init_options: Optional[Dict[str, Any]]=None,
                 trainable: bool=True) -> None:
        """
        Constructor of class QuarticBSpline integrating the C++/CUDA extension
        'quartic_bspline_extension'.

        :param num_marginals: Number of potentials - must coincide with number of filters
        :param box_lower: Lower bound of spline domain, defaults to -1.0
        :param box_upper: Upper bound of spline domain, defaults to 1.0
        :param num_centers: _description_, defaults to 87
        :param init_options: _description_, defaults to None
        :param trainable: _description_, defaults to True
        """
        super().__init__(num_marginals)

        default_init = {
            'mode': 'student_t', 
            'multiplier': 1.0, 
        }
        self.init_options = {**default_init, **(init_options or {})}

        self.box_lower = box_lower
        self.box_upper = box_upper
        self.num_centers = num_centers
        self.scale = (self.box_upper - self.box_lower) / (self.num_centers - 1)
        self.register_buffer('centers', torch.linspace(self.box_lower, self.box_upper, self.num_centers))

        weight_data = self._init_weight_tensor(num_marginals, self.centers, self.init_options)
        self.weight_tensor = torch.nn.Parameter(data=weight_data, requires_grad=trainable)

    @staticmethod
    def _init_weight_tensor(num_marginals: int, centers: torch.Tensor, init_options: Dict[str, Any]) -> torch.Tensor:
        multiplier = init_options['multiplier']
        if init_options['mode'] == 'rand':
            weight_data = torch.log(multiplier * torch.rand(num_marginals, len(centers)))
        elif init_options['mode'] == 'uniform':
            weight_data = torch.log(multiplier * torch.ones(num_marginals, len(centers)))
        elif init_options['mode'] == 'student_t':
            weight_data = torch.log(multiplier * torch.log(1 + torch.stack([centers 
                                                                for _ in range(0, num_marginals)], dim=0) ** 2))
        else:
            raise ValueError('Unknown initialisation method')
        return weight_data
        
    def initialisation_dict(self) -> Dict[str, Any]:
        return {'num_marginals': self.num_marginals, 
                'box_lower': self.box_lower,
                'box_upper': self.box_upper,
                'num_centers': self.num_centers}

    def get_parameters(self) -> torch.Tensor:
        return self.weight_tensor.data

    def freeze(self) -> None:
        self.weight_tensor.requires_grad = False
    
    def unfreeze(self) -> None:
        self.weight_tensor.requires_grad = True

    def forward(self, x: torch.Tensor, reduce: bool=True) -> torch.Tensor:
        y, _ = QuarticBSplineFunction.apply(2 * torch.arctan(x) / torch.pi, torch.exp(self.weight_tensor), self.centers, self.scale)
        if reduce:
            return torch.sum(y)
        else:
            return y

    @classmethod
    def from_file(cls, path_to_model: str, device: torch.device=torch.device('cpu')) -> Self:
        potential_data = torch.load(path_to_model, map_location=device)

        initialisation_dict = potential_data['initialisation_dict']
        state_dict = potential_data['state_dict']

        num_marginals = initialisation_dict.get('num_marginals', 48)
        box_lower = initialisation_dict.get('box_lower', -1.0)
        box_upper = initialisation_dict.get('box_upper', 1.0)
        num_centers = initialisation_dict.get('num_centers', 87)
        potential = cls(num_marginals=num_marginals, box_lower=box_lower, box_upper=box_upper, num_centers=num_centers)
        potential.load_state_dict(state_dict, strict=True)
        return potential

    @classmethod
    def from_config(cls, config: Configuration) -> Self:
        num_marginals = config['potential']['quartic_bspline']['num_marginals'].get()
        box_lower = config['potential']['quartic_bspline']['box_lower'].get()
        box_upper = config['potential']['quartic_bspline']['box_upper'].get()
        num_centers = config['potential']['quartic_bspline']['num_centers'].get()
        initialisation_mode = config['potential']['quartic_bspline']['initialisation']['mode'].get()
        multiplier = config['potential']['quartic_bspline']['initialisation']['multiplier'].get()
        trainable = config['potential']['quartic_bspline']['trainable'].get()
        
        init_options = {'mode': initialisation_mode, 'multiplier': multiplier}

        return cls(num_marginals=num_marginals, 
                   box_lower=box_lower, 
                   box_upper=box_upper, 
                   num_centers=num_centers,
                   init_options=init_options,
                   trainable=trainable)

    def save(self, path_to_model_dir: str, model_name: str='quartic_spline') -> str:
        potential_dict = {'initialisation_dict': self.initialisation_dict(),
                          'state_dict': self.state_dict()}
        return Potential.save(potential_dict, path_to_model_dir, model_name)




import os
import logging
import torch

from pylopt.data.LogLevel import LogLevel

def setup_logger(data_dir_path: str = '.', log_level_str: str = 'info') -> None:
    log_level = LogLevel.from_string(log_level_str)
    logging.basicConfig(level=log_level,
                        format='[{asctime}.{msecs:0<3.0f}][{levelname}] {message}',
                        style='{', datefmt='%Y-%m-%dT%H:%M:%S',
                        handlers=[logging.StreamHandler(),
                                  logging.FileHandler(os.path.join(data_dir_path, 'application.log'))])

def log_trainable_params_stats(model: torch.nn.Module, logging_module: str) -> None:
    num_params_total = sum([p.numel() for p in model.parameters()])
    num_params_trainable = sum([p.numel() for p in model.parameters() if p.requires_grad])

    logging.info('[{:s}] parameter stats'.format(logging_module.upper()))
    logging.info('[{:s}]   > total number of parameters: {:d}'.format(logging_module.upper(), num_params_total))
    logging.info('[{:s}]   > trainable parameters: {:d}'.format(logging_module.upper(), num_params_trainable))


from typing import List, Dict, Any

from pylopt.optimise.optimise_adam import LR_KEY

DEFAULTS_GROUP_LBFGS = {LR_KEY: 1, 'max_iter': 20, 'max_eval': 25, 'tolerance_grad': 1e-7,
                        'tolerance_change': 1e-9, 'history_size': 20, 'line_search_fn': None}

def harmonise_param_groups_lbfgs(param_groups: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    param_groups_ = []
    for group in param_groups:
        group_ = {'params': [p for p in group['params']],
                  'history': [p.detach().clone().requires_grad_(True) for p in group['params']]}
        for key in group.keys():
            if key != 'params':
                group_[key] = group[key]

        for key in DEFAULTS_GROUP_LBFGS.keys():
            group_[key] = group[key] if group[key] is not None else DEFAULTS_GROUP_LBFGS[key]

        param_groups_.append(group_)
    return param_groups_


from typing import List
import torch
from torchvision.transforms import v2
import logging

def collate_function(batch: List[torch.Tensor], crop_size: int = 64) -> torch.Tensor:
    if crop_size > 0:
        return torch.cat([v2.RandomCrop(size=crop_size)(item) for item in batch], dim=0)
    else:
        min_width = min([u.shape[-1] for u in batch])
        max_width = max([u.shape[-1] for u in batch])
        min_height = min([u.shape[-2] for u in batch])
        max_height = max([u.shape[-2] for u in batch])

        if (min_width < max_width) or (min_height < max_height):
            logging.info('[COLLATE] Batch contains images of different shapes: apply padding.')
            batch_padded = []
            for i in range(0, len(batch)):
                u = batch[i]
                diff_horizontal = max_height - u.shape[-1]
                diff_vertical = max_width - u.shape[-2]

                u_padded = torch.nn.functional.pad(batch[i], (0, diff_horizontal, 0, diff_vertical),
                                                   mode='constant', value=1)
                batch_padded.append(u_padded)
            ret_val = torch.cat([item for item in batch_padded], dim=0)
        else:
            ret_val = torch.cat([item for item in batch], dim=0)

        return ret_val


import torch

def compute_psnr(y_true: torch.Tensor, y_pred: torch.Tensor, max_pix_value: float = 1.0) -> torch.Tensor:
    mse = torch.mean(((y_true - y_pred) ** 2), dim=(-2, -1))
    return 20 * torch.log10(max_pix_value / torch.sqrt(mse))



import torch
from typing import Optional

from pylopt.regularisers.fields_of_experts.ImageFilter import ImageFilter
from pylopt.regularisers.fields_of_experts.potential.Potential import Potential

PARAM_GROUP_NAME_KEY = 'name'
PARAM_GROUP_NAME_DICT = {Potential.__name__: 'potentials', ImageFilter.__name__: 'filters'}

def get_param_group_name(module: torch.nn.Module) -> Optional[str]:
    group_name = None
    for cls in type(module).__mro__:
        if cls.__name__ in PARAM_GROUP_NAME_DICT.keys():
            group_name = PARAM_GROUP_NAME_DICT[cls.__name__]
    return group_name


import os
from confuse import Configuration
from pathlib import Path
from typing import Dict, Any, Optional
import yaml

def dump_configs(config: Configuration, path_to_data_dir: str) -> None:
    config_data = config.dump(full=True)
    config_file_path = os.path.join(path_to_data_dir, 'config.yaml')
    with open(config_file_path, 'w') as file:
        file.write(str(config_data))

def dump_bilevel_training_settings(upper_settings: Dict[str, Any], lower_settings: Dict[str, Any],
                                   backward_settings: Dict[str, Any],  path_to_data_dir: str) -> None:
    settings_dict = {'upper_problem': upper_settings, 'lower_problem': lower_settings, 'backward': backward_settings}
    settings_file_path = os.path.join(path_to_data_dir, 'bilevel_training_settings.yaml')
    with open(settings_file_path, 'w') as file:
        yaml.dump(settings_dict, file, sort_keys=False)

def create_experiment_dir(config: Configuration) -> str:
    experiments_root_dir = config['data']['experiments']['root_dir'].get()
    if not experiments_root_dir:
        package_root_path = Path(__file__).resolve().parents[2]
        experiments_root_dir = os.path.join(package_root_path, 'data', 'evaluation')
        os.makedirs(experiments_root_dir, exist_ok=True)

    experiment_list = sorted(os.listdir(experiments_root_dir))
    if experiment_list:
        experiment_id = str(int(experiment_list[-1]) + 1).zfill(5)
    else:
        experiment_id = str(0).zfill(5)
    path_to_eval_dir = os.path.join(experiments_root_dir, experiment_id)
    os.makedirs(path_to_eval_dir, exist_ok=True)

    return path_to_eval_dir

def get_repo_root_path(start: Path = Path(__file__).resolve(), marker: str='pyproject.toml') -> Optional[Path]:
    repo_root_path = None
    for parent in [start, *start.parents]:
        if (parent / marker).exists():
            repo_root_path = parent
            break
    
    return repo_root_path


import os.path
from typing import Dict, Any, Self, Optional
import torch
from confuse import Configuration

from . import Potential

class StudentT(Potential):
    """
    Class implementing student-t potential for the usage in context of FoE models.
    """
    def __init__(
            self,
            num_marginals: int,
            init_options: Optional[Dict[str, Any]]=None,
            trainable: bool=True
    ) -> None:
        """
        Initialisation of class StudentT

        NOTE
        ----
            > It is not checked whether number of filters and number of potentials coincide.

        :param num_marginals: Number of potentials - per design it must coincide
            with the number of filters.
        :param init_options: Dictionary of initialisation options. Keys:
            - 'mode': String indicating how potentials weights are initialised: 'rand', 'uniform'
            - 'multiplier': Float used as initial scaling factor
            Defaults to None
        :param trainable: Boolean flag indicating if potentials are trainable; defaults to True
        """
        super().__init__(num_marginals)

        default_init = {
            'mode': 'uniform', 
            'multiplier': 1.0
        }
        self.init_options = {**default_init, **(init_options or {})}

        weight_data = self._init_weight_tensor(num_marginals, self.init_options)
        self.weight_tensor = torch.nn.Parameter(data=weight_data, requires_grad=trainable)

    @staticmethod
    def _init_weight_tensor(num_marginals: int, init_options: Dict[str, Any]) -> torch.Tensor:
        multiplier = init_options['multiplier']

        if init_options['mode'] == 'rand':
            weight_data = torch.log(multiplier * torch.rand(num_marginals))
        elif init_options['mode'] == 'uniform':
            weight_data = torch.log(multiplier * torch.ones(num_marginals))
        else:
            raise ValueError('Unknown initialisation method')
        return weight_data

    def get_parameters(self) -> torch.Tensor:
        return self.weight_tensor.data

    def freeze(self) -> None:
        self.weight_tensor.requires_grad = False
    
    def unfreeze(self) -> None:
        self.weight_tensor.requires_grad = True

    def forward(self, x: torch.Tensor, reduce: bool=True) -> torch.Tensor:
        if reduce:
            return torch.einsum('bfhw,f->', torch.log(1.0 + x ** 2), torch.exp(self.weight_tensor))
        else:
            return torch.einsum('bfhw,f->bfhw', torch.log(1.0 + x ** 2), torch.exp(self.weight_tensor))

    def initialisation_dict(self) -> Dict[str, Any]:
        return {'num_marginals': self.num_marginals}

    @classmethod
    def from_file(cls, path_to_model: str, device: torch.device=torch.device('cpu')) -> Self:
        potential_data = torch.load(path_to_model, map_location=device)

        initialisation_dict = potential_data['initialisation_dict']
        state_dict = potential_data['state_dict']

        num_marginals = initialisation_dict.get('num_marginals', 48)
        potential = cls(num_marginals=num_marginals)
        potential.load_state_dict(state_dict, strict=True)
        return potential

    @classmethod
    def from_config(cls, config: Configuration) -> Self:
        num_marginals = config['potential']['student_t']['num_marginals'].get()
        initialisation_mode = config['potential']['student_t']['initialisation']['mode'].get()
        multiplier = config['potential']['student_t']['initialisation']['multiplier'].get()
        trainable = config['potential']['student_t']['trainable'].get()

        init_options = {'mode': initialisation_mode, 'multiplier': multiplier}
        return cls(num_marginals, init_options, trainable)

    def save(self, path_to_model_dir: str, model_name: str='student_t') -> str:
        potential_dict = {'initialisation_dict': self.initialisation_dict(),
                          'state_dict': self.state_dict()}
        return Potential.save(potential_dict, path_to_model_dir, model_name)


from abc import ABC, abstractmethod
from typing import List, Dict, Any, Optional, Callable
from math import cos, pi

from pylopt.bilevel_problem.parameter_groups import PARAM_GROUP_NAME_KEY
from pylopt.optimise import LIP_CONST_KEY

class HyperParamScheduler(ABC):

    def __init__(self):
        self.step_counter = 0

    @abstractmethod
    def bind(self, param_groups: List[Dict[str, Any]]) -> None:
        pass

    @abstractmethod
    def step(self, **kwargs: Dict[str, Any]) -> None:
        pass

class AdaptiveLRRestartScheduler(HyperParamScheduler):
    """
    Class implementing adaptive restarts handled by means of the callable condition_func, which
    takes kwargs as input only, and must return a boolean value indicating if a restart shall
    be performed.

    NOTE
    ----
        > The authors of [1] indicate several suitable restart policies
            * Fixed interval restart
            * Adaptive restart
                - Loss-based: Restart the scheme if over a predefined number of steps (patience) the
                    loss does not decrease
                - Gradient-based: Restart if grad * (x_{k} - x_{k - 1}) > 0 over a predefined number of
                    steps (patience), i.e. current search direction is not a descent direction.

    References
    ----------
    [1] O’donoghue, B. and Candes, E., 2015. Adaptive restart for accelerated gradient schemes.
        Foundations of computational mathematics, 15(3), pp.715-732.
    """

    def __init__(self, 
                 condition_func: Callable, 
                 gamma: float=0.9999, 
                 warm_up_period: int=100, 
                 patience: int=10,
                 lr_key: str='lr'
    ) -> None:
        super().__init__()

        self.condition_func = condition_func
        self.gamma = gamma
        self.warm_up_period = warm_up_period
        self.patience = patience
        self.lr_key = lr_key

        self.num_bad_iterations = 0

        self.param_groups = None
        self.base_values = None

    def bind(self, param_groups: List[Dict[str, Any]]) -> None:
        self.param_groups = param_groups
        self.base_values = []
        for group in param_groups:
            self.base_values.append({self.lr_key: group.get(self.lr_key, None)})

    def _update_param_group(self) -> None:
        for idx, group in enumerate(self.param_groups):
            if self.lr_key in group.keys():
                group[self.lr_key] *= self.gamma

    def _perform_restart(self) -> None:
        for idx, group in enumerate(self.param_groups):
            if self.lr_key in group.keys() and self.base_values[idx][self.lr_key] is not None:
                group[self.lr_key] = self.base_values[idx][self.lr_key]

    def step(self, **kwargs: Dict[str, Any]) -> None:
        self.step_counter += 1

        if self.condition_func(self.param_groups, **kwargs) and self.step_counter >= self.warm_up_period:
            self.num_bad_iterations += 1
            if self.num_bad_iterations == self.patience:
                self._perform_restart()
                self.num_bad_iterations = 0

        self._update_param_group()

class CosineAnnealingLRScheduler(HyperParamScheduler):
    """
    Implementation of cosine annealing lr scheduler with optional fixed interval restarts. Annealing is applied
    only within iteration window which needs to be specified.

    NOTE
    ----
        > This scheduler is particularly suited for optimisers from torch.optim like Adam, SGD, ...
        > Updates are performed for the learning rate of all (!) parameter groups.
    """
    def __init__(self, step_begin: int, step_end: int, lr_min: float=1e-7, lr_key: str='lr',
                 restart_cycle: Optional[int]=None, restart_cycle_multi: Optional[float]=None) -> None:
        """
        Initialisation of cosine annealing lr scheduler.

        :param step_begin: Begin of annealing window
        :param step_end: End of annealing window
        :param lr_min: Minimal learning rate
        :param lr_key: Key of learning within parameter groups
        :param restart_cycle: Optional integer indicating after how many iterations (within the annealing window)
            a restart shall be performed; if None no restarts were performed.
        :param restart_cycle_multi: Multiplier to adjust length of the restart cycle over training; if None
            the restart cycles keep the same length.
        """
        super().__init__()

        self.step_begin = step_begin
        self.step_end = step_end
        self.lr_min = lr_min
        self.lr_key = lr_key
        self.decay_cycle_len = restart_cycle if restart_cycle is not None else self.step_end - self.step_begin
        self.decay_cycle_multi = 1 if (restart_cycle is None or restart_cycle_multi is None) \
            else restart_cycle_multi

        self.steps_since_restart_counter = 0

        self.param_groups = None
        self.base_values = None

    def bind(self, param_groups: List[Dict[str, Any]]) -> None:
        self.param_groups = param_groups
        self.base_values = []
        for group in param_groups:
            self.base_values.append({self.lr_key: group.get(self.lr_key, None)})

    def _update_param_group(self):
        for idx, group in enumerate(self.param_groups):
            if self.lr_key in group.keys() and self.base_values[idx][self.lr_key] is not None:
                group[self.lr_key] = (self.lr_min +
                                      0.5 * (self.base_values[idx][self.lr_key] - self.lr_min) *
                                      (1 + cos((self.steps_since_restart_counter / self.decay_cycle_len) * pi)))

    def step(self,  **kwargs: Dict[str, Any]) -> None:
        self.step_counter += 1
        if self.step_begin <= self.step_counter < self.step_end:
            if self.steps_since_restart_counter > 0 and self.steps_since_restart_counter % self.decay_cycle_len == 0:
                self.steps_since_restart_counter = -1
                self.decay_cycle_len = int(self.decay_cycle_multi * self.decay_cycle_len)
            self.steps_since_restart_counter += 1
            self._update_param_group()

class AdaptiveNAGRestartScheduler(HyperParamScheduler):

    def __init__(self, 
                 condition_func: Callable, 
                 warm_up_period: int=100, 
                 patience: int=10,
                 lip_const_key: str=LIP_CONST_KEY,
                 beta_key: str='beta',              # momentum parameter
                 theta_key: str='theta'             # parameter use to compute momentum parameter
    ) -> None:
        super().__init__()

        self.condition_func = condition_func
        self.lip_const_key = lip_const_key

        self.param_groups = None
        self.base_values = None

    def bind(self, param_groups: List[Dict[str, Any]]) -> None:
        self.param_groups = param_groups

        self.base_values = []
        for group in param_groups:
            self.base_values.append({self.lip_const_key: group.get(self.lip_const_key, None)})

    def _perform_restart() -> None:

        # reset lip_const
        # refresh history
        # reset momentum parameters
        for idx, group in enumerate(self.param_groups):
            pass

    def step(self, **kwargs) -> None:
        self.step_counter += 1

        if self.condition_func(self.param_groups, **kwargs) and self.step_counter >= self.warm_up_period:
            self.num_bad_iterations += 1
            if self.num_bad_iterations == self.patience:
                self._perform_restart()
                self.num_bad_iterations = 0

        # if self.step_counter % self.restart_freq == 0:
        #     for idx, group in enumerate(self.param_groups):
        #         for key in self.base_values[idx]:
        #             if key in group.keys():
        #                 group[key] = self.base_values[idx].get(key, group[key])

class NAGLipConstGuard(HyperParamScheduler):
    """
    Scheduler for NAG Lipschitz constant. If the Lipschitz constant of all parameter groups or a specified parameter
    group is larger than a specified bound, it will set to 0.5 times this upper bound. Per default all parameter
    groups will be affected.
    """
    def __init__(self, lip_const_bound: float, lip_const_key: str=LIP_CONST_KEY,
                 group_name: Optional[str]=None) -> None:
        """
        Initialisation of Lipschitz constant guard

        :param lip_const_bound: Upper bound of Lipschitz constant
        :param lip_const_key: Key of Lipschitz constant within parameter group.
        :param group_name:
        """
        super().__init__()

        self.lip_const_bound = lip_const_bound
        self.lip_const_key = lip_const_key
        self.group_name = group_name if group_name is not None else 'all'

        self.param_groups = None

    def bind(self, param_groups: List[Dict[str, Any]]) -> None:
        self.param_groups = param_groups

    def _find_param_group(self, group_name: str) -> Optional[Dict[str, Any]]:
        group = None
        for grp in self.param_groups:
            if hasattr(grp, PARAM_GROUP_NAME_KEY) and getattr(grp, PARAM_GROUP_NAME_KEY) == group_name:
                group = grp
        return group

    def _update_lip_const(self, group: Dict[str, Any]) -> Dict[str, Any]:
        """
        Function which updates the Lipschitz constant of a parameter group if the group does
        contain the corresponding key (self.lip_const_key). Otherwise, the group and its parameters
        won't be altered.

        :param group: Parameter group in terms of dictionary
        :return: Updated parameter group.
        """
        if self.lip_const_key in group.keys():
            group[self.lip_const_key] = [min(0.5 * self.lip_const_bound, item) for item in group[self.lip_const_key]]
        return group

    def step(self, **kwargs: Dict[str, Any]) -> None:
        self.step_counter += 1

        if self.group_name == 'all':
            for group in self.param_groups:
                self._update_lip_const(group)
        else:
            group = self._find_param_group(self.group_name)
            if group is None:
                raise ValueError('There is no parameter group with name {:s}'.format(self.group_name))
            self._update_lip_const(group)



import torch
import os
from matplotlib import pyplot as plt
import cv2

from bilevel_optimisation.utils.config_utils import load_app_config
from bilevel_optimisation.fields_of_experts import FieldsOfExperts
from bilevel_optimisation.filters import ImageFilter
from bilevel_optimisation.potential import Potential


path_to_data_dir = '/home/florianthaler/Documents/research/bilevel_optimisation/bilevel_optimisation/model_data/'

        # filters = torch.load(os.path.join(path_to_data_dir, 'foe_filters_7x7_chen-ranftl-pock_2014_scaled.pt'))
        # potentials = torch.load(os.path.join(path_to_data_dir, 'foe_thetas_7x7_chen-ranftl-pock_2014.pt'))

img_1 = cv2.imread('/home/florianthaler/Documents/data/image_data/some_images/watercastle.jpg', cv2.IMREAD_GRAYSCALE) / 255
img_2 = cv2.imread('/home/florianthaler/Documents/data/image_data/some_images/giraffe.jpg', cv2.IMREAD_GRAYSCALE) / 255

img_1 = torch.from_numpy(img_1).unsqueeze(dim=0).unsqueeze(dim=0).to(dtype=torch.float32)
img_2 = torch.from_numpy(img_2).unsqueeze(dim=0).unsqueeze(dim=0).to(dtype=torch.float32)
img = torch.cat([img_1, img_2], dim=0)

config = load_app_config('app_name', 'example_prediction_I', '')
image_filter = ImageFilter(config)
potential = Potential.from_config(image_filter.get_num_filters(), config)
regulariser = FieldsOfExperts(potential, image_filter)

image_filter.to(dtype=torch.float32)
regulariser.to(dtype=torch.float32)

y = image_filter(img)
y = torch.permute(y, dims=(1, 0, 2, 3))
y_flat = torch.flatten(y, start_dim=1, end_dim=3)

quantiles = torch.quantile(y_flat, q=torch.tensor([0.025, 0.975]), dim=-1)



num_marginals = 48
num_marginals_sqrt = 7

# filter_norms = torch.sum(filters['state_dict']['filter_tensor'] ** 2, dim=(-3, -2, -1)).unsqueeze(dim=0).unsqueeze(dim=2).unsqueeze(dim=3).cpu()





fig, axes = plt.subplots(num_marginals_sqrt, num_marginals_sqrt, figsize=(13, 13),
                         gridspec_kw={'hspace': 0.9, 'wspace': 0.4}, sharex=False, sharey=False)
for i in range(0, num_marginals_sqrt):
    for j in range(0, num_marginals_sqrt):
        potential_idx = i * num_marginals_sqrt + j
        if potential_idx < num_marginals:

            q_low = quantiles[0, potential_idx]
            q_high = quantiles[1, potential_idx]

            t = torch.stack([torch.linspace(q_low, q_high, 111)
                             for _ in range(0, num_marginals)]).unsqueeze(dim=1).unsqueeze(dim=0)
            y = torch.log(1 + t ** 2)



            axes[i, j].plot(t[0, potential_idx, 0, :].detach().cpu().numpy(),
                            y[0, potential_idx, 0, :].detach().cpu().numpy() -
                            torch.min(y[0, potential_idx, 0, :]).detach().cpu().numpy(), color='blue')
            # potential_weight_tensor = potentials['state_dict']['weight_tensor']
            # potential_weight = potential_weight_tensor[potential_idx].detach().cpu().item()
            # axes[i, j].set_title('idx={:d}, \nweight={:.3f}'.format(potential_idx, potential_weight),
            #                      fontsize=8)
            axes[i, j].set_xlim(q_low, q_high)

            for label in axes[i, j].get_xticklabels():
                label.set_fontsize(8)

            for label in axes[i, j].get_yticklabels():
                label.set_fontsize(8)

        else:
            fig.delaxes(axes[i, j])

plt.show()




import random
import numpy as np
import torch

def seed_random_number_generators(seed_val: int) -> None:
    torch.manual_seed(seed_val)
    torch.cuda.manual_seed(seed_val)
    np.random.seed(seed_val)
    random.seed(seed_val)



import torch
from abc import ABC
from typing import Dict, Any

from pylopt.regularisers.fields_of_experts import FieldsOfExperts, compile_regulariser
from pylopt.energy import MeasurementModel

class Energy(torch.nn.Module, ABC):
    """
    Class, inheriting from torch.nn.Module which represents the energy function assumed to be the sum of
    a data fidelty term and a regularisation term:

        (1 / (2 * sigma**2)) * | u - u_{noisy}| ** 2 + lam * regulariser(u)

    """
    def __init__(self, measurement_model: MeasurementModel, regulariser: FieldsOfExperts, lam: float) -> None:
        """
        Initialisation of an object of class InnerEnergy

        :param measurement_model: Object of class MeasurementModel
        :param regulariser: Fields of experts regulariser
        :param lam: Regularisation parameter
        """
        super().__init__()
        self.measurement_model = measurement_model
        self.regulariser = regulariser
        self.lam = lam

    def compile(self, **compile_options: Dict[str, Any]) -> None:
        """
        Compilation of a regulariser using torch.compile().

        NOTE
        ----
            > Per default the most aggressive compilation options are used.
            > [IMPORTANT] When model shall be applied to different image shapes, then use the option dynamic=True.
                By default, this option is applied. Additionally, note that several warm up calls are performed
                using tensors of shape measurement_model.get_noisy_observation().shape.

        :param compile_options: Optional compilation options
        :return:
        """
        self.regulariser = compile_regulariser(self.regulariser,
                                               self.measurement_model.get_noisy_observation(), **compile_options)

    def forward(self, u: torch.Tensor) -> torch.Tensor:
        """
        Returns sum of data fidelty and (scaled) regularisation term

        :param u: Tensor at which data fidelty and regulariser are evaluated
        :return: Tensor representing the energy at the input x
        """
        return self.measurement_model(u) + self.lam * self.regulariser(u)



import torch

from bilevel_optimisation.potential import NaturalCubicSpline
from bilevel_optimisation.utils.config_utils import load_app_config

def main():

    path_to_config_dir = '/home/florianthaler/Documents/research/bilevel_optimisation/bilevel_optimisation/config_data/custom/example_prediction_III'

    config = load_app_config('test', path_to_config_dir, 'test_module')
    spline = NaturalCubicSpline(48, config)

    t = torch.stack([torch.linspace(-2, 2, 111)
                     for _ in range(0, 48)]).unsqueeze(dim=1).unsqueeze(dim=0).requires_grad_(True)
    y = spline.forward_negative_log(t, reduce=False)
    y1 = spline.first_derivative(t, grad_outputs=torch.ones_like(t))
    y2 = spline.second_derivative(t, grad_outputs=torch.ones_like(t), mixed=False, retain_graph=True)
    y3 = spline.second_derivative(t, grad_outputs=torch.ones_like(t), mixed=True)

    from matplotlib import pyplot as plt
    num_marginals_sqrt = 7

    fig, axes = plt.subplots(num_marginals_sqrt, num_marginals_sqrt, figsize=(11, 11),
                             gridspec_kw={'hspace': 0.9, 'wspace': 0.2}, sharex=True, sharey=True)
    for i in range(0, num_marginals_sqrt):
        for j in range(0, num_marginals_sqrt):
            potential_idx = i * num_marginals_sqrt + j
            if potential_idx < spline.get_num_marginals():
                idx = potential_idx

                axes[i, j].plot(t[0, potential_idx, 0, :].detach().cpu().numpy(),
                                y[0, potential_idx, 0, :].detach().cpu().numpy() -
                                torch.min(y[0, potential_idx, 0, :]).detach().cpu().numpy(), color='blue')

                axes[i, j].set_title('idx={:d}'.format(idx), fontsize=8)

                axes[i, j].set_xlim(-2, 2)
            else:
                fig.delaxes(axes[i, j])

    plt.show()

if __name__ == '__main__':
    main()




import torch
from typing import Optional, List, Dict, Any, Tuple, Callable

from pylopt.data import LowerProblemResult, OptimiserResult
from pylopt.energy import Energy
from pylopt.energy import MeasurementModel
from pylopt.optimise import optimise_nag, optimise_adam, optimise_nag_unrolling, LIP_CONST_KEY

def make_prox_map(prox_operator: torch.nn.Module, u: torch.Tensor) -> Callable:
    def prox_map(x: torch.Tensor, tau: float) -> torch.Tensor:
        return prox_operator(x, tau, u)
    return prox_map

def add_prox(prox_operator: Optional[torch.nn.Module], param_groups: List[Dict[str, Any]], u: torch.Tensor) -> None:
    u_ = u.detach().clone()
    prox_map = make_prox_map(prox_operator, u_)
    for group in param_groups:
        for p in group['params']:
            setattr(p, 'prox', prox_map)

def assemble_param_groups_nag(u: torch.Tensor, alpha: Optional[float]=None,
                              beta: Optional[float]=None, lip_const: Optional[float]=None,
                              batch_optimisation: bool=True, **unknown_options) -> List[Dict[str, Any]]:
    u_ = u.detach().clone()
    group = {'params': [torch.nn.Parameter(u_, requires_grad=True)]}

    num_optimisation_variables = 1 if batch_optimisation else u.shape[0]
    group['alpha'] = [alpha] * num_optimisation_variables if alpha is not None else [None] * num_optimisation_variables
    group['beta'] = [beta] * num_optimisation_variables if beta is not None else [None] * num_optimisation_variables
    group[LIP_CONST_KEY] = [lip_const] * num_optimisation_variables if lip_const is not None \
        else [None] * num_optimisation_variables

    return [group]

def assemble_param_groups_adam(u: torch.Tensor, lr: Optional[float]=None,
                               betas: Optional[Tuple[float, float]]=None,
                               weight_decay: Optional[float]=None,
                               batch_optimisation: bool=True, **unknown_options) -> List[Dict[str, Any]]:
    u_ = u.detach().clone()
    param_groups = []
    if batch_optimisation:
        group = {'params': [torch.nn.Parameter(u_, requires_grad=True)],
                 'lr': lr, 'betas': betas, 'weight_decay': weight_decay}
        param_groups.append(group)
    else:
        for i in range(0, u_.shape[0]):
            group = {'params': [torch.nn.Parameter(u_[i: i + 1, :, :, :].detach().clone(), requires_grad=True)],
                     'lr': lr, 'betas': betas, 'weight_decay': weight_decay}
            param_groups.append(group)
    return param_groups

def parse_result(result: OptimiserResult, max_num_iterations: int, **unknown_options) -> LowerProblemResult:
    solution_tensor = build_solution_tensor_from_param_groups(result.solution)
    lower_prob_result = LowerProblemResult(solution=solution_tensor, num_iterations=result.num_iterations,
                                           loss=result.loss,
                                           message='Converged' if result.num_iterations < max_num_iterations
                                           else 'Max. number of iterations reached')
    return lower_prob_result

def build_solution_tensor_from_param_groups(param_groups: List[Dict[str, Any]]):
    solution_list = [group['params'][-1] for group in param_groups]
    return torch.cat(solution_list, dim=0)

def build_objective_func(energy: Energy, 
                         batch_optim: bool, 
                         use_prox: bool
) -> Callable:
    if batch_optim and not use_prox:
        def func(x: torch.Tensor) -> torch.Tensor:
            return energy(x)
    elif batch_optim and use_prox:
        def func(x: torch.Tensor) -> torch.Tensor:
            return energy.lam * energy.regulariser(x)
    elif not batch_optim and not use_prox:
        operator = energy.measurement_model.operator
        noise_level = energy.measurement_model.noise_level

        regulariser = energy.regulariser
        lam = energy.lam

        per_sample_energy_models = []
        u_clean = energy.measurement_model.u_clean
        for i in range(u_clean.shape[0]):
            sample_measurement_model = MeasurementModel(u_clean[i:i + 1, :, :, :], operator=operator,
                                                        noise_level=noise_level)
            sample_energy = Energy(sample_measurement_model, regulariser, lam)
            per_sample_energy_models.append(sample_energy)

        # RECALL
        #   > torch.vmap cannot be used for vectorisation of a list of different maps.
        def func(*x: torch.Tensor) -> torch.Tensor:
            return torch.stack([sample_energy_model(sample.unsqueeze(dim=0)) for sample_energy_model, sample in
                                zip(per_sample_energy_models, torch.cat(x, dim=0))])

        return func
    else:
        regulariser = energy.regulariser
        lam = energy.lam
        def func(x: torch.Tensor) -> torch.Tensor:            
            return lam * torch.sum(regulariser(x, reduce=False), dim=(-3, -2, -1))

    return func

def build_gradient_func(func: Callable, batch_optim: bool, unrolling: bool=False) -> Callable:
    if batch_optim:
        def grad_func(x: torch.Tensor) -> List[torch.Tensor]:
            with torch.enable_grad():
                x_ = x.detach().clone().requires_grad_(True)
                loss = func(x_)
            return list(torch.autograd.grad(outputs=loss, inputs=[x_], create_graph=unrolling))
    else:
        def grad_func(*x: torch.Tensor):
            with torch.enable_grad():
                x = [x_.detach().clone().requires_grad_(True) for x_ in x]
                loss = torch.sum(func(*x), dim=0)
            return list(torch.autograd.grad(outputs=loss, inputs=x, create_graph=unrolling))

    return grad_func

def solve_lower(energy: Energy, method: str, options: Dict[str, Any]) -> LowerProblemResult:
    """
    :param energy: Instance of class Energy which represents the lower problem.
    :param method: String indicating which of the provided optimisation methods is used to solve the lower problem.
        The following methods are implemented:
        - 'nag' : Nestorov's accelerated gradient method.
        - 'napg': Proximal gradient method with Nesterov acceleration.
        - 'adam': Adam optimisation scheme
    :param options: Options need to be provided in terms of a dictionary. The options required depend on the chosen
        method.
        - method == 'nag'
          ---------------
          max_num_iterations: int
            Maximal number of iterations to perform
          rel_tol: float, optional
            Tolerance for early stopping of the optimisation scheme. If not provided, max_num_iterations iterations
            will be performed.
          batch_optimisation: bool, optional
            Flag indicating if batch-optimisation or per-sample-optimisation is performed. Batch-optimisation performs
            gradient steps on the full batch. Hence, only a single step size or a single Lipschitz constant, and a
            single momentum parameter are required. Per-sample-optimisation performs for each sample of the batch an
            independent gradient step. Consequently, each sample requires step size or Lipschitz constant, and a
            momentum parameter.
          alpha: float, optional
            Constant step size used in the optimisation scheme. If not specified, backtracking line search
            to obtain sufficient descent along the direction of the negative gradient is applied.
            If alpha is specified and batch_optimisation is set to False, the specified alpha is used for each sample.
          beta: float, optional
            Constant momentum parameter. If not specified, the momentum parameter (for the full batch
            and for each sample respectively) is computed at iteration level k as follows

                theta_{k + 1} = 0.5 * (1 + math.sqrt(1 + 4 * (theta_{k} ** 2)))
                beta_k = (theta_{k} - 1) / theta_{k + 1}

            If beta is specified, and batch_optimisation == False, the specified value is used for all the samples
            in the optimisation scheme.
          lip_const: float, optional
            Initial Lipschitz constant used within the backtracking line search. The specified value is used in case
            of batch-optimisation as well as in case of per-sample optimisation.
            If lip_const and alpha are not specified, the default value 1e5 for the full batch optimisation or
            per sample optimisation is used.
          resample_measurement_noise: bool, optional
            Boolean flag indicating if measurement noise is resampled at each iteration. 
        - method == 'napg'
          ---------------
          max_num_iterations: int
            As for 'nag'
          rel_tol: float, optional
            As for 'nag'
          batch_optimisation: bool, optional
            As for 'nag'
          alpha: float, optional
            As for 'nag'
          beta: float, optional
            As for 'nag'
          lip_const: float, optional
            As for 'nag'
          prox: torch.nn.Module
            A PyTorch module which represents the proximal operator. Its forward function must take the arguments
              u: torch.Tensor representing the item at which the operator shall be evaluated
              tau: torch.Tensor representing the step size of the gradient step
              u_noisy: torch.Tensor representing the noisy observation
        - method == 'adam'
          ---------------
          max_num_iterations: int
            As for 'nag'
          batch_optimisation: bool, optional
            This flag indicates, similarly as for 'nag', 'napg', if each sample of the batch should be treated
            individually within the optimisation process. If set to True, for each sample separate hyperparameters (lr,
            betas, weight_decay) are used. Note that per-sample optimisation is organised in this context by means
            of different parameter groups, which in PyTorch are processed sequentially.
          lr: float, optional
            Learning rate, which is used for the single parameter group in batch-optimisation and for all the
            parameter groups in case of per-sample optimisation. Per default, 1e-4 is used if
            no specifications were made.
          betas: Tuple[float, float], optional
            Parameter tuple used to compute moving average of first and second moments of gradients. The same handling
            as for lr is applied. Per default the tuple (0.9, 0.999) is used.
          weight_decay: float, optional
            Weights for L^2 penalisation. Again, the same handling as for 'lr' is applied. Per default the
            value 0.0 is used.
          resample_measurement_noise: bool, optional
            Boolean flag indicating if measurement noise is resampled at each iteration. 
    :return: Instance of class LowerProblemResult containing the solution tensor, the number of performed iterations,
        the final loss, and a message indicating why iteration terminated.
    """
    batch_optim = options.get('batch_optimisation', True)
    resample_measurement_noise = options.get('resample_measurement_noise', False)
    energy.measurement_model.resample_measurement_noise = resample_measurement_noise

    if method == 'nag':
        func = build_objective_func(energy, 
                                    batch_optim=batch_optim, 
                                    use_prox=False)
        grad_func = build_gradient_func(func, batch_optim=batch_optim)

        u_noisy = energy.measurement_model.get_noisy_observation().detach().clone()
        param_groups = assemble_param_groups_nag(u_noisy, **options)

        nag_result = optimise_nag(func, grad_func, param_groups, **options)
        lower_prob_result = parse_result(nag_result, **options)
    elif method == 'napg':
        func = build_objective_func(energy, batch_optim=batch_optim, use_prox=True)
        grad_func = build_gradient_func(func, batch_optim=batch_optim)

        u_noisy = energy.measurement_model.get_noisy_observation().detach().clone()
        param_groups = assemble_param_groups_nag(u_noisy, **options)

        prox_operator = options.get('prox', None)
        add_prox(prox_operator, param_groups, u_noisy)

        nag_result = optimise_nag(func, grad_func, param_groups, **options)
        lower_prob_result = parse_result(nag_result, **options)
    elif method == 'adam':
        func_ = build_objective_func(energy, 
                                     batch_optim=batch_optim, 
                                     use_prox=False)
        func = lambda *z: torch.sum(func_(torch.cat(z, dim=0)))

        u_noisy = energy.measurement_model.get_noisy_observation().detach().clone()
        param_groups = assemble_param_groups_adam(u_noisy, **options)
        
        adam_result = optimise_adam(func, param_groups, **options)
        lower_prob_result = parse_result(adam_result, **options)
    elif method == 'nag_unrolling':
        func = build_objective_func(energy, batch_optim=batch_optim, use_prox=False)
        grad_func = build_gradient_func(func, batch_optim, unrolling=True)

        u = energy.measurement_model.get_noisy_observation().detach().clone()
        param_groups = assemble_param_groups_nag(u.requires_grad_(True), **options)

        nag_result = optimise_nag_unrolling(func, grad_func, param_groups, **options)
        lower_prob_result = parse_result(nag_result, **options)
    elif method == 'napg_unrolling':
        func = build_objective_func(energy, batch_optim=batch_optim, use_prox=False)
        grad_func = build_gradient_func(func, batch_optim, unrolling=True)

        u = energy.measurement_model.get_noisy_observation().detach().clone()
        param_groups = assemble_param_groups_nag(u.requires_grad_(True), **options)

        prox_operator = options.get('prox', None)
        add_prox(prox_operator, param_groups, u)

        nag_result = optimise_nag_unrolling(func, grad_func, param_groups, **options)
        lower_prob_result = parse_result(nag_result, **options)
    
    elif method == 'your_custom_method':
        # Custom method for solving the lower problem or by map
        # estimation of by mmse estimation goes here. Use the following
        # structure
        #
        #       solution = solution_method(...)
        #       lower_prob_result = LowerProblemResult(...)
        pass
    else:
        raise ValueError('Unknown solution method for lower level problem.')

    return lower_prob_result





import torch
from torch.utils.data import DataLoader
import os
from matplotlib import pyplot as plt
from matplotlib import colormaps as cmaps
import numpy as np
from pathlib import Path
import pandas as pd
from scipy import ndimage

from pylopt.dataset.dataset_utils import collate_function
from pylopt.dataset.ImageDataset import TestImageDataset
from pylopt.regularisers.fields_of_experts.potential.StudentT import StudentT
from pylopt.regularisers.fields_of_experts.ImageFilter import ImageFilter
from pylopt.utils.file_system_utils import get_repo_root_path


def normalise_filter(filter_tensor: torch.Tensor) -> torch.Tensor:
    filter_tensor = filter_tensor - torch.min(filter_tensor)
    filter_tensor = filter_tensor / torch.max(filter_tensor)
    return filter_tensor

data_root_path = '/home/florianthaler/Documents/trash/00053'
filter_model_path = 'models/filters_iter_14999.pt'
potential_model_path = 'models/potential_iter_14999.pt'
model = torch.load(os.path.join(data_root_path, filter_model_path))

filter_tensor = model['state_dict']['filter_tensor']

filter_norms = [torch.linalg.norm(fltr).detach().cpu().item() for fltr in filter_tensor]

filter_indices_sorted = np.argsort(filter_norms)[::-1].tolist()

num_filters = filter_tensor.shape[0]
n = 3
m = 16
fig1, axes = plt.subplots(n, m, figsize=(7, 2), gridspec_kw={'hspace': 0.3, 'wspace': 0.3})
for i in range(0, n):
    for j in range(0, m):
        filter_idx = i * m + j
        if filter_idx < num_filters:
            idx = filter_indices_sorted[filter_idx]

            fltr = normalise_filter(filter_tensor[idx, :, :, :].squeeze().detach().clone())
            axes[i, j].imshow(fltr.cpu().numpy(), cmap=cmaps['gray'])

            # title = 'idx={:d}, \nnorm={:.3f}'.format(idx, filter_norms[idx])
            # axes[i, j].set_title(title, fontsize=8)
            axes[i, j].axis('off')
        else:
            fig1.delaxes(axes[i, j])

fig2, axes = plt.subplots(n, m, figsize=(7, 2), gridspec_kw={'hspace': 0.3, 'wspace': 0.2}, sharey=True)

fltr = ImageFilter.from_file(os.path.join(data_root_path, filter_model_path)).to(device=torch.device('cpu'))
pot = StudentT.from_file(os.path.join(data_root_path, potential_model_path)).to(device=torch.device('cpu'))

root_path = get_repo_root_path(Path(__file__))
test_data_root_dir = os.path.join(root_path, 'data', 'images', 'test_images')
test_image_dataset = TestImageDataset(root_path=test_data_root_dir, dtype=torch.float32)
test_loader = DataLoader(test_image_dataset, 
                            batch_size=len(test_image_dataset), 
                            shuffle=False,
                            collate_fn=lambda x: collate_function(x, crop_size=-1))

u_clean = list(test_loader)[0]

filter_response = fltr(u_clean)
filter_response = torch.permute(filter_response, dims=(1, 0, 2, 3))
filter_response_flat = torch.flatten(filter_response, start_dim=1, end_dim=3)
quantiles = torch.quantile(filter_response_flat,
                            q=torch.tensor([0.025, 1 - 0.025]).to(device=torch.device('cpu')), dim=-1)

t = torch.stack([torch.linspace(quantiles[0, i], quantiles[1, i], 111)
                    for i in range(0, 48)]).unsqueeze(dim=1).unsqueeze(dim=0)
t = t.to(device=torch.device('cpu'), dtype=torch.float32)

rho = pot.forward(t, reduce=False)


for i in range(0, n):
    for j in range(0, m):
        potential_idx = i * m + j
        if potential_idx < pot.get_num_marginals():
            idx = filter_indices_sorted[potential_idx]

            axes[i, j].plot(t[0, idx, 0, :].detach().cpu().numpy(),
                            rho[0, idx, 0, :].detach().cpu().numpy() -
                            torch.min(rho[0, potential_idx, 0, :]).detach().cpu().numpy(), color='blue')

            potential_weight_tensor = pot.get_parameters()
            potential_weight = potential_weight_tensor[idx].detach().cpu().item()
            # axes[i, j].set_title('idx={:d}, \nweight={:.3f}'.format(idx, potential_weight),
            #                         fontsize=8)

            # axes[i, j].set_xlim(quantiles[0, idx].cpu().item(), quantiles[1, idx].cpu().item())
            axes[i, j].set_xticks([])
            axes[i, j].set_xticks([], minor=True)

            axes[i, j].set_yticks([])
            axes[i, j].set_yticks([], minor=True)

            # for x_label, y_label in zip(axes[i, j].get_xticklabels(), axes[i, j].get_yticklabels()):
            #     x_label.set_fontsize(8)
            #     y_label.set_fontsize(8)
        else:
            fig2.delaxes(axes[i, j])


# plt.show()


df = pd.read_csv('~/Downloads/csv.csv')
df = df[df['Step'] <= 10000]


raw = df['Value']
smoothed = ndimage.uniform_filter1d(df['Value'], size=10)

import seaborn as sns
sns.set_style("whitegrid") 

fig3 = plt.figure()
ax = fig3.add_subplot(1, 1, 1)
ax.plot(df['Step'], raw, label='raw')
ax.plot(df['Step'], smoothed, label='smoothed')
ax.grid(True)
ax.set_xlabel('iteration')
ax.set_ylabel('psnr [dB]')
ax.legend(loc='lower right')

# plt.style.use('seaborn-v0_8')  # or 'seaborn' for older versions
# plt.figure(figsize=(12, 8))
# plt.plot(df['Step'], df['Value'], linewidth=1.5)
# plt.grid(True, alpha=0.3)

plt.show()






from typing import Any, List, Dict
import torch
from torch.autograd import Function
from torch.autograd.function import FunctionCtx

from pylopt.energy.Energy import Energy
from pylopt.solver import LinearSystemSolver
from pylopt.lower_problem.solve_lower import solve_lower

def compute_hvp_state(energy: Energy, u: torch.Tensor, v: torch.Tensor) -> torch.Tensor:
    """
    Function which is used to compute the Hessian vector product of the lower level energy function w.r.t.
    the state variable u.

    NOTE
    ----
        > The computation exploits the special structure of the energy function

                (1 / (2 * sigma**2)) * | u - u_{noisy}| ** 2 + lam * regulariser(u)

            and computes the second derivative simply as the sum of the exact second derivative
            of the data fidelty term, and the second derivative of the regulariser using its
            method second_derivative(). This allows to use autograd whenever needed, or
            to circumvent autograd if analytical expressions for the derivatives are known (as
            for splines).

    :param energy: PyTorch module representing the lower level energy funxtion
    :param u: Tensor at which second order derivative shall be computed
    :param v: Tensor at which the Hessian is applied
    :return: Tensor representing the result of the Hessian at u, applied to v.
    """
    with torch.enable_grad():
        x = u.detach().clone()
        x.requires_grad = True

        e = energy(x)
        de_dx = torch.autograd.grad(inputs=x, outputs=e, create_graph=True)
    return torch.autograd.grad(inputs=x, outputs=de_dx[0], grad_outputs=v)[0]

def compute_hvp_mixed(energy: Energy, u: torch.Tensor, v: torch.Tensor) -> List[torch.Tensor]:
    """
    Function for computing the mixed second order derivative of the energy function.

    :param energy: PyTorch module representing the lower level energy
    :param u: Tensor at which derivatives need to be computed.
    :param v: Tensor to which derivatives are applied.
    :return: Tensor representing the result of mixed Hessian at u, applied to v.
    """
    with torch.enable_grad():
        x = u.detach().clone()
        x.requires_grad = True

        e = energy(x)
        de_dx = torch.autograd.grad(inputs=x, outputs=e, create_graph=True)
    d2e_mixed = torch.autograd.grad(inputs=[p for p in energy.parameters() if p.requires_grad],
                                    outputs=de_dx, grad_outputs=v)
    return list(d2e_mixed)

class ImplicitAutogradFunction(Function):
    """
    Subclass of torch.autograd.Function. It implements the implicit differentiation scheme
    to compute the gradients of optimiser of the inner problem w.r.t. to the parameters
    of the regulariser - as references for implicit differentiation see [1], [2]. For the
    custom implementation of the backward call see [3].

    References
    ----------
    [1] Chen, Y., Ranftl, R. and Pock, T., 2014. Insights into analysis operator learning: From patch-based
        sparse models to higher order MRFs. IEEE Transactions on Image Processing, 23(3), pp.1060-1072.
    [2] Samuel, K.G. and Tappen, M.F., 2009, June. Learning optimized MAP estimates in continuously-valued
        MRF models. In 2009 IEEE Conference on Computer Vision and Pattern Recognition (pp. 477-484). IEEE.
    [3] https://docs.pytorch.org/docs/stable/notes/extending.html
    """

    @staticmethod
    def forward(ctx: FunctionCtx, energy: Energy, method_lower: str, options_lower: Dict[str, Any],
                loss_func: torch.nn.Module, solver: LinearSystemSolver, hessian_free: bool,
                *params: torch.nn.Parameter) -> torch.Tensor:
        """
        Function which needs to be implemented due to subclassing from torch.autograd.Function.
        It computes and provides data which is required in the backward step.

        :param ctx:
        :param energy: PyTorch module representing the energy function of the lower level problem.
        :param method_lower: String indicating which method is used to solve the lower level problem.
        :param options_lower: Dictionary of options regarding the optimisation method for the lower
            level problem.
        :param loss_func: PyTorch module representing the upper loss function
        :param solver: Linear system solver of class LinearSystemSolver
        :param hessian_free: Boolean indicating if 0-th order approximation of the Hessian of the energy function
            is used for the gradient computation.
        :param params: List of PyTorch parameters whose gradients need to be computed.
        :return: Current upper loss
        """
        u_denoised = solve_lower(energy, method_lower, options_lower).solution
        ctx.save_for_backward(u_denoised.detach().clone())

        ctx.energy = energy
        ctx.loss_func = loss_func
        ctx.solver = solver
        ctx.hessian_free = hessian_free

        return loss_func(u_denoised)

    @staticmethod
    def compute_lagrange_multiplier(outer_loss_func: torch.nn.Module, energy: Energy,
                                    u_denoised: torch.Tensor, solver: LinearSystemSolver,
                                    hessian_free: bool=False) -> torch.Tensor:
        """
        Function which computes the Lagrange multiplier of the KKT-formulation of the bilevel problem.

        :param outer_loss_func: PyTorch module representing the outer loss
        :param energy: PyTorch module representing the inner problem
        :param u_denoised: Result of denoising procedure
        :param solver: Linear system solver
        :param hessian_free: Bool indicating if 0-th order approximation of the operator of the linear
            system shall be used or not. Per default the value is set to False
        :return: Solution of linear system in terms of a PyTorch tensor.
        """
        with torch.enable_grad():
            x = u_denoised.detach().clone()
            x.requires_grad = True
            outer_loss = outer_loss_func(x)
        grad_outer_loss = torch.autograd.grad(outputs=outer_loss, inputs=x)[0]

        if hessian_free:
            return -grad_outer_loss
        else:
            lin_operator = lambda z: compute_hvp_state(energy, u_denoised, z)
            lagrange_multiplier_result = solver.solve(lin_operator, -grad_outer_loss)

            return lagrange_multiplier_result.solution

    @staticmethod
    def backward(ctx: FunctionCtx, *grad_output: torch.Tensor) -> Any:
        """
        This function implements the custom backward step based on the principle of implicit differentiation.

        References
        ----------
        [1] https://docs.pytorch.org/docs/stable/notes/extending.html

        :param ctx: This kind of object is used to pass tensors, and other objects from the forward call
            to the backward call. For more details see [1]
        :param grad_output: Not used in this implementation
        :return: Gradients of outer loss w.r.t. the parameters of regulariser. For each input of the
            forward function there must be a return parameter. For more details see again [1].
        """
        u_denoised = ctx.saved_tensors[0]
        energy = ctx.energy
        outer_loss_func = ctx.loss_func
        solver = ctx.solver
        hessian_free = ctx.hessian_free
        lagrange_multiplier = ImplicitAutogradFunction.compute_lagrange_multiplier(outer_loss_func, energy,
                                                                                   u_denoised, solver,
                                                                                   hessian_free=hessian_free)
        grad_params = compute_hvp_mixed(energy, u_denoised.detach(), lagrange_multiplier)

        energy.zero_grad()

        return None, None, None, None, None, None, *grad_params

class UnrollingAutogradFunction(Function):
    """
    Subclass of torch.autograd.Function with the purpose to provide a custom backward
    function based on an unrolling scheme.
    """
    @staticmethod
    def forward(ctx: FunctionCtx, energy: Energy, method_lower: str, options_lower: Dict[str, Any],
                loss_func: torch.nn.Module, *params) -> torch.Tensor:
        """

        :param ctx:
        :param energy:
        :param method_lower:
        :param options_lower:
        :param loss_func:
        :param params:
        :return:
        """
        with torch.enable_grad():
            u_denoised = solve_lower(energy, method_lower, options_lower).solution
            loss = loss_func(u_denoised)
        grad_params = torch.autograd.grad(outputs=loss, inputs=params)

        ctx.grad_params = grad_params
        ctx.energy = energy
        return loss

    @staticmethod
    def backward(ctx: FunctionCtx, *grad_output: torch.Tensor) -> Any:
        """

        :param ctx:
        :param grad_output:
        :return:
        """
        grad_params = ctx.grad_params
        energy = ctx.energy
        energy.zero_grad()

        return None, None, None, None, *grad_params


import logging
from enum import Enum

class LogLevel(Enum):
    NOTSET = logging.NOTSET
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL

    @staticmethod
    def from_string(log_level_str: str) -> int:
        try:
            return LogLevel[log_level_str.upper()].value
        except KeyError:
            raise ValueError('Invalid log level {:s}'.format(log_level_str))




import os
import pathlib
import torch
from confuse import Configuration

from pylopt.filters import ImageFilter

def equal(filter_0: ImageFilter, filter_1:ImageFilter) -> bool:
    return (filter_0.filter_dim == filter_1.filter_dim
            and filter_0.padding == filter_1.padding
            and filter_0.padding_mode == filter_1.padding_mode
            and filter_0.orthogonality_options == filter_1.orthogonality_options
            and torch.allclose(filter_0.filter_tensor, filter_1.filter_tensor))

def test_save_and_load(tmp_path: pathlib.Path):
    dtype = torch.float64

    image_filter_0 = ImageFilter().to(dtype=dtype)
    image_filter_0.save(str(tmp_path), 'image_filter.pt')

    model_path = os.path.join(tmp_path, 'image_filter.pt')
    image_filter_1 = ImageFilter.from_file(model_path).to(dtype=dtype)

    assert equal(image_filter_0, image_filter_1)

def test_load_from_config():
    dtype = torch.float64

    config = Configuration('unit_test')
    config.set_file(os.path.join('./tests/data', 'test_config.yaml'))

    torch.manual_seed(123)
    image_filter_0 = ImageFilter(
        filter_dim=config['image_filter']['filter_dim'].get(),
        padding=config['image_filter']['padding'].get(),
        padding_mode=config['image_filter']['padding_mode'].get(),
        initialisation_mode=config['image_filter']['initialisation']['mode'].get(),
        multiplier=config['image_filter']['initialisation']['multiplier'].get(),
        normalise=config['image_filter']['initialisation']['normalise'].get(),
        trainable=config['image_filter']['trainable'].get(),
        orthogonality_options={
            'enable': config['image_filter']['orthogonality']['enable'].get(),
            'max_num_iterations': config['image_filter']['orthogonality']['max_num_iterations'].get()
        }
    ).to(dtype=dtype)

    torch.manual_seed(123)
    image_filter_1 = ImageFilter.from_config(config).to(dtype=dtype)

    assert equal(image_filter_0, image_filter_1)



import torch

def solve_tridiagonal(diag: torch.Tensor, diag_super: torch.Tensor, diag_sub: torch.Tensor,
                      rhs: torch.Tensor) -> torch.Tensor:
    """
    This function solves a tridiagonal linear system given a single system matrix, and multiple right hand sides.
    For the solution, the Thomas algorithm is applied - which corresponds basically to the application of
    Gaussian elimination while exploiting the special structure of the system matrix

    NOTE
    ----
        > Regularity of the system matrix is not checked. It is assumed that the system matrix
            is indeed regular.
        > The input tensors diag, rhs will NOTE (!) be altered - altering these would lower memory consumption though.

    :param diag: PyTorch tensor representing the (main) diagonal of the system matrix
    :param diag_super: PyTorch tensor representing the super diagonal of the system matrix
    :param diag_sub: PyTorch tensor representing the sub diagonal of the system matrix
    :param rhs: Two-dimensional PyTorch tensor, where dimension 0 is assumed to be the batch dimension; i.e. if
        the system has n unknowns, and it shall be solved for m right hand sides, the tensor rhs is assumed to
        be of shape [m, n].
    :return: PyTorch tensor representing the solution tensor. It has the same shape as the right hand side tensor
        rhs.
    """
    diag_ = torch.clone(diag)
    rhs_ = torch.clone(rhs)

    n = rhs_.shape[1]
    for i in range(1, n):
        tmp = diag_sub[i - 1] / diag_[i - 1]
        diag_[i] = diag_[i] - tmp * diag_super[i - 1]
        rhs_[:, i] = rhs_[:, i] - tmp * rhs_[:, i - 1]

    x = torch.empty_like(rhs)
    x[:, n - 1] = rhs_[:, -1] / diag_[-1]
    for i in range(n - 2, -1, -1):
        x[:, i] = (rhs_[:, i] - diag_super[i] * x[:, i + 1]) / diag_[i]
    return x


import torch
from dataclasses import dataclass
from typing import Optional, Dict, Any

@dataclass
class LowerProblemResult:
    solution: torch.Tensor
    num_iterations: int
    loss: torch.Tensor
    message: str=''
    meta: Optional[Dict[str, Any]]=None
